@@ -1,4 +1,6 @@
 import logging
+import math
+import os
 import random
 import re
 from types import SimpleNamespace
@@ -10,21 +12,28 @@ import torch.nn.functional as F
 
 import stanza.models.classifiers.classifier_args as classifier_args
 import stanza.models.classifiers.data as data
-from stanza.models.common.vocab import PAD_ID, UNK_ID
+from stanza.models.common.bert_embedding import extract_bert_embeddings
 from stanza.models.common.data import get_long_tensor, sort_all
+from stanza.models.common.foundation_cache import load_bert
 from stanza.models.common.utils import split_into_batches, sort_with_indices, unsort
-# TODO: move CharVocab to common
-from stanza.models.pos.vocab import CharVocab
+from stanza.models.common.vocab import PAD_ID, UNK_ID
 
 """
 The CNN classifier is based on Yoon Kim's work:
 
 https://arxiv.org/abs/1408.5882
 
+Also included are maxpool 2d, conv 2d, and a bilstm, as in
+
+Text Classification Improved by Integrating Bidirectional LSTM
+with Two-dimensional Max Pooling
+https://aclanthology.org/C16-1329.pdf
+
 The architecture is simple:
 
 - Embedding at the bottom layer
   - separate learnable entry for UNK, since many of the embeddings we have use 0 for UNK
+- maybe a bilstm layer, as per a command line flag
 - Some number of conv2d layers over the embedding
 - Maxpool layers over small windows, window size being a parameter
 - FC layer to the classification layer
@@ -39,10 +48,11 @@ help, but dev performance went down for each variation of
 """
 
 logger = logging.getLogger('stanza')
+tlogger = logging.getLogger('stanza.classifiers.trainer')
 
 class CNNClassifier(nn.Module):
     def __init__(self, pretrain, extra_vocab, labels,
-                 charmodel_forward, charmodel_backward,
+                 charmodel_forward, charmodel_backward, elmo_model, bert_model, bert_tokenizer,
                  args):
         """
         pretrain is a pretrained word embedding.  should have .emb and .vocab
@@ -72,6 +82,12 @@ class CNNClassifier(nn.Module):
                                       extra_wordvec_max_norm = args.extra_wordvec_max_norm,
                                       char_lowercase = args.char_lowercase,
                                       charlm_projection = args.charlm_projection,
+                                      use_elmo = args.use_elmo,
+                                      elmo_projection = args.elmo_projection,
+                                      bert_model = args.bert_model,
+                                      bilstm = args.bilstm,
+                                      bilstm_hidden_dim = args.bilstm_hidden_dim,
+                                      maxpool_width = args.maxpool_width,
                                       model_type = 'CNNClassifier')
 
         self.char_lowercase = args.char_lowercase
@@ -80,17 +96,23 @@ class CNNClassifier(nn.Module):
 
         emb_matrix = pretrain.emb
         self.add_unsaved_module('embedding', nn.Embedding.from_pretrained(torch.from_numpy(emb_matrix), freeze=True))
+        self.add_unsaved_module('elmo_model', elmo_model)
         self.vocab_size = emb_matrix.shape[0]
         self.embedding_dim = emb_matrix.shape[1]
 
-        self.add_unsaved_module('charmodel_forward', charmodel_forward)
+        self.add_unsaved_module('forward_charlm', charmodel_forward)
         if charmodel_forward is not None:
-            self.add_unsaved_module('charmodel_forward_vocab', charmodel_forward.char_vocab())
-            logger.debug("Got forward char model of dimension {}".format(charmodel_forward.hidden_dim()))
-        self.add_unsaved_module('charmodel_backward', charmodel_backward)
+            tlogger.debug("Got forward char model of dimension {}".format(charmodel_forward.hidden_dim()))
+            if not charmodel_forward.is_forward_lm:
+                raise ValueError("Got a backward charlm as a forward charlm!")
+        self.add_unsaved_module('backward_charlm', charmodel_backward)
         if charmodel_backward is not None:
-            self.add_unsaved_module('charmodel_backward_vocab', charmodel_backward.char_vocab())
-            logger.debug("Got backward char model of dimension {}".format(charmodel_backward.hidden_dim()))
+            tlogger.debug("Got backward char model of dimension {}".format(charmodel_backward.hidden_dim()))
+            if charmodel_backward.is_forward_lm:
+                raise ValueError("Got a forward charlm as a backward charlm!")
+
+        self.add_unsaved_module('bert_model', bert_model)
+        self.add_unsaved_module('bert_tokenizer', bert_tokenizer)
 
         # The Pretrain has PAD and UNK already (indices 0 and 1), but we
         # possibly want to train UNK while freezing the rest of the embedding
@@ -98,7 +120,8 @@ class CNNClassifier(nn.Module):
         # you want to spend a long time debugging this
         self.unk = nn.Parameter(torch.randn(self.embedding_dim) / np.sqrt(self.embedding_dim) / 10.0)
 
-        self.vocab_map = { word: i for i, word in enumerate(pretrain.vocab) }
+        # replacing NBSP picks up a whole bunch of words for VI
+        self.vocab_map = { word.replace('\xa0', ' '): i for i, word in enumerate(pretrain.vocab) }
 
         if self.config.extra_wordvec_method is not classifier_args.ExtraVectors.NONE:
             if not extra_vocab:
@@ -118,7 +141,7 @@ class CNNClassifier(nn.Module):
                                                 embedding_dim = self.config.extra_wordvec_dim,
                                                 max_norm = self.config.extra_wordvec_max_norm,
                                                 padding_idx = 0)
-            logger.debug("Extra embedding size: {}".format(self.extra_embedding.weight.shape))
+            tlogger.debug("Extra embedding size: {}".format(self.extra_embedding.weight.shape))
         else:
             self.extra_vocab = None
             self.extra_vocab_map = None
@@ -152,60 +175,119 @@ class CNNClassifier(nn.Module):
                 self.charmodel_backward_projection = None
                 total_embedding_dim += charmodel_backward.hidden_dim()
 
-        self.conv_layers = nn.ModuleList([nn.Conv2d(in_channels=1,
-                                                    out_channels=self.config.filter_channels,
-                                                    kernel_size=(filter_size, total_embedding_dim))
-                                          for filter_size in self.config.filter_sizes])
+        if self.config.use_elmo:
+            if elmo_model is None:
+                raise ValueError("Model requires elmo, but elmo_model not passed in")
+            elmo_dim = elmo_model.sents2elmo([["Test"]])[0].shape[1]
 
-        previous_layer_size = len(self.config.filter_sizes) * self.config.filter_channels
+            # this mapping will combine 3 layers of elmo to 1 layer of features
+            self.elmo_combine_layers = nn.Linear(in_features=3, out_features=1, bias=False)
+            if self.config.elmo_projection:
+                self.elmo_projection = nn.Linear(in_features=elmo_dim, out_features=self.config.elmo_projection)
+                total_embedding_dim = total_embedding_dim + self.config.elmo_projection
+            else:
+                total_embedding_dim = total_embedding_dim + elmo_dim
+
+        if bert_model is not None:
+            if bert_tokenizer is None:
+                raise ValueError("Cannot have a bert model without a tokenizer")
+            self.bert_dim = self.bert_model.config.hidden_size
+            total_embedding_dim += self.bert_dim
+
+        if self.config.bilstm:
+            conv_input_dim = self.config.bilstm_hidden_dim * 2
+            self.bilstm = nn.LSTM(batch_first=True,
+                                  input_size=total_embedding_dim,
+                                  hidden_size=self.config.bilstm_hidden_dim,
+                                  num_layers=2,
+                                  bidirectional=True,
+                                  dropout=0.2)
+        else:
+            conv_input_dim = total_embedding_dim
+            self.bilstm = None
+
+        self.fc_input_size = 0
+        self.conv_layers = nn.ModuleList()
+        self.max_window = 0
+        for filter_idx, filter_size in enumerate(self.config.filter_sizes):
+            if isinstance(filter_size, int):
+                self.max_window = max(self.max_window, filter_size)
+                if isinstance(self.config.filter_channels, int):
+                    filter_channels = self.config.filter_channels
+                else:
+                    filter_channels = self.config.filter_channels[filter_idx]
+                fc_delta = filter_channels // self.config.maxpool_width
+                tlogger.debug("Adding full width filter %d.  Output channels: %d -> %d", filter_size, filter_channels, fc_delta)
+                self.fc_input_size += fc_delta
+                self.conv_layers.append(nn.Conv2d(in_channels=1,
+                                                  out_channels=filter_channels,
+                                                  kernel_size=(filter_size, conv_input_dim)))
+            elif isinstance(filter_size, tuple) and len(filter_size) == 2:
+                filter_height, filter_width = filter_size
+                self.max_window = max(self.max_window, filter_width)
+                if isinstance(self.config.filter_channels, int):
+                    filter_channels = max(1, self.config.filter_channels // (conv_input_dim // filter_width))
+                else:
+                    filter_channels = self.config.filter_channels[filter_idx]
+                fc_delta = filter_channels * (conv_input_dim // filter_width) // self.config.maxpool_width
+                tlogger.debug("Adding filter %s.  Output channels: %d -> %d", filter_size, filter_channels, fc_delta)
+                self.fc_input_size += fc_delta
+                self.conv_layers.append(nn.Conv2d(in_channels=1,
+                                                  out_channels=filter_channels,
+                                                  stride=(1, filter_width),
+                                                  kernel_size=(filter_height, filter_width)))
+            else:
+                raise ValueError("Expected int or 2d tuple for conv size")
+
+        tlogger.debug("Input dim to FC layers: %d", self.fc_input_size)
         fc_layers = []
+        previous_layer_size = self.fc_input_size
         for shape in self.config.fc_shapes:
             fc_layers.append(nn.Linear(previous_layer_size, shape))
             previous_layer_size = shape
         fc_layers.append(nn.Linear(previous_layer_size, self.config.num_classes))
         self.fc_layers = nn.ModuleList(fc_layers)
 
-        self.max_window = max(self.config.filter_sizes)
-
         self.dropout = nn.Dropout(self.config.dropout)
 
     def add_unsaved_module(self, name, module):
         self.unsaved_modules += [name]
         setattr(self, name, module)
 
-    def build_char_reps(self, batch_chars, batch_offsets, device, forward=True):
-        if forward:
-            model = self.charmodel_forward
-            vocab = self.charmodel_forward_vocab
-            projection = self.charmodel_forward_projection
-        else:
-            model = self.charmodel_backward
-            vocab = self.charmodel_backward_vocab
-            projection = self.charmodel_backward_projection
-
-        batch_charlens = [len(x) for x in batch_chars]
-        chars_sorted, char_orig_idx = sort_all([batch_chars, batch_offsets], batch_charlens)
-        batch_chars, batch_offsets = chars_sorted
-        batch_charlens = [len(x) for x in batch_chars]
-        chars = get_long_tensor(batch_chars, len(batch_chars),
-                                pad_id=vocab.unit2id(' ')).to(device=device)
-        char_reps = model.get_representation(chars, batch_offsets,
-                                             batch_charlens, char_orig_idx)
-        char_reps = char_reps.data
+    def build_char_reps(self, inputs, max_phrase_len, charlm, projection, begin_paddings, device):
+        char_reps = charlm.build_char_representation(inputs)
         if projection is not None:
-            char_reps = projection(char_reps)
-        char_reps = torch.reshape(char_reps, [max(len(x) for x in batch_offsets), len(batch_chars), char_reps.shape[-1]])
-        char_reps = torch.transpose(char_reps, 0, 1)
+            char_reps = [projection(x) for x in char_reps]
+        char_inputs = torch.zeros((len(inputs), max_phrase_len, char_reps[0].shape[-1]), device=device)
+        for idx, rep in enumerate(char_reps):
+            start = begin_paddings[idx]
+            end = start + rep.shape[0]
+            char_inputs[idx, start:end, :] = rep
+        return char_inputs
 
-        return char_reps
+    def extract_bert_embeddings(self, inputs, max_phrase_len, begin_paddings, device):
+        bert_embeddings = extract_bert_embeddings(self.config.bert_model, self.bert_tokenizer, self.bert_model, inputs, device, keep_endpoints=False)
+        bert_inputs = torch.zeros((len(inputs), max_phrase_len, bert_embeddings[0].shape[-1]), device=device)
+        for idx, rep in enumerate(bert_embeddings):
+            start = begin_paddings[idx]
+            end = start + rep.shape[0]
+            bert_inputs[idx, start:end, :] = rep
+        return bert_inputs
 
-    def char_case(self, x: str) -> str:
-        return x.lower() if self.char_lowercase else x
+    def forward(self, inputs):
+        # assume all pieces are on the same device
+        device = next(self.parameters()).device
 
-    def forward(self, inputs, device=None):
-        if not device:
-            # assume all pieces are on the same device
-            device = next(self.parameters()).device
+        vocab_map = self.vocab_map
+        def map_word(word):
+            idx = vocab_map.get(word, None)
+            if idx is not None:
+                return idx
+            if word[-1] == "'":
+                idx = vocab_map.get(word[:-1], None)
+                if idx is not None:
+                    return idx
+            return vocab_map.get(word.lower(), UNK_ID)
 
         # we will pad each phrase so either it matches the longest
         # conv or the longest phrase in the input, whichever is longer
@@ -216,10 +298,11 @@ class CNNClassifier(nn.Module):
         batch_indices = []
         batch_unknowns = []
         extra_batch_indices = []
-        batch_forward_chars = []
-        batch_forward_offsets = []
-        batch_backward_chars = []
-        batch_backward_offsets = []
+        begin_paddings = []
+        end_paddings = []
+
+        elmo_batch_words = []
+
         for phrase in inputs:
             # we use random at training time to try to learn different
             # positions of padding.  at test time, though, we want to
@@ -230,35 +313,19 @@ class CNNClassifier(nn.Module):
                 begin_pad_width = 0
             end_pad_width = max_phrase_len - begin_pad_width - len(phrase)
 
+            begin_paddings.append(begin_pad_width)
+            end_paddings.append(end_pad_width)
+
             # the initial lists are the length of the begin padding
             sentence_indices = [PAD_ID] * begin_pad_width
+            sentence_indices.extend([map_word(x) for x in phrase])
+            sentence_indices.extend([PAD_ID] * end_pad_width)
+
             # the "unknowns" will be the locations of the unknown words.
             # these locations will get the specially trained unknown vector
-            sentence_unknowns = []
+            # TODO: split UNK based on part of speech?  might be an interesting experiment
+            sentence_unknowns = [idx for idx, word in enumerate(sentence_indices) if word == UNK_ID]
 
-            for word in phrase:
-                if word in self.vocab_map:
-                    sentence_indices.append(self.vocab_map[word])
-                    continue
-                new_word = word.replace("-", "")
-                # google vectors have words which are all dashes
-                if len(new_word) == 0:
-                    new_word = word
-                if new_word in self.vocab_map:
-                    sentence_indices.append(self.vocab_map[new_word])
-                    continue
-
-                if new_word[-1] == "'":
-                    new_word = new_word[:-1]
-                    if new_word in self.vocab_map:
-                        sentence_indices.append(self.vocab_map[new_word])
-                        continue
-
-                # TODO: split UNK based on part of speech?  might be an interesting experiment
-                sentence_unknowns.append(len(sentence_indices))
-                sentence_indices.append(PAD_ID)
-
-            sentence_indices.extend([PAD_ID] * end_pad_width)
             batch_indices.append(sentence_indices)
             batch_unknowns.append(sentence_unknowns)
 
@@ -284,34 +351,12 @@ class CNNClassifier(nn.Module):
                 extra_sentence_indices.extend([PAD_ID] * end_pad_width)
                 extra_batch_indices.append(extra_sentence_indices)
 
-            if self.charmodel_forward is not None:
-                start_id, end_id = self.charmodel_forward_vocab.unit2id('\n'), self.charmodel_forward_vocab.unit2id(' ') # special token
-                processed_phrase = ([[end_id]] * begin_pad_width +
-                                    [self.charmodel_forward_vocab.map([self.char_case(x) for x in w]) + [end_id] for w in phrase] +
-                                    [[end_id]] * end_pad_width)
-                chars = [start_id]
-                offsets = []
-                for word in processed_phrase:
-                    # -1 because we want the position of the last character
-                    chars.extend(word)
-                    offsets.append(len(chars)-1)
-                batch_forward_chars.append(chars)
-                batch_forward_offsets.append(offsets)
-
-            if self.charmodel_backward is not None:
-                start_id, end_id = self.charmodel_backward_vocab.unit2id('\n'), self.charmodel_backward_vocab.unit2id(' ') # special token
-                processed_phrase = ([[end_id]] * begin_pad_width +
-                                    [self.charmodel_backward_vocab.map([self.char_case(x) for x in w[::-1]]) + [end_id] for w in phrase[::-1]] +
-                                    [[end_id]] * end_pad_width)
-                chars = [start_id]
-                offsets = []
-                for word in processed_phrase:
-                    # -1 because we want the position of the last character
-                    chars.extend(word)
-                    # backwards because that seems to be what get_representation wants
-                    offsets = [len(chars)-1] + offsets
-                batch_backward_chars.append(chars)
-                batch_backward_offsets.append(offsets)
+            if self.config.use_elmo:
+                elmo_phrase_words = [""] * begin_pad_width
+                for word in phrase:
+                    elmo_phrase_words.append(word)
+                elmo_phrase_words.extend([""] * end_pad_width)
+                elmo_batch_words.append(elmo_phrase_words)
 
         # creating a single large list with all the indices lets us
         # create a single tensor, which is much faster than creating
@@ -329,8 +374,7 @@ class CNNClassifier(nn.Module):
         # we use the random unk so that we are not necessarily
         # learning to match 0s for unk
         for phrase_num, sentence_unknowns in enumerate(batch_unknowns):
-            for unknown in sentence_unknowns:
-                input_vectors[phrase_num, unknown, :] = self.unk
+            input_vectors[phrase_num][sentence_unknowns] = self.unk
 
         if self.extra_vocab:
             extra_batch_indices = torch.tensor(extra_batch_indices, requires_grad=False, device=device)
@@ -344,25 +388,56 @@ class CNNClassifier(nn.Module):
         else:
             all_inputs = [input_vectors]
 
-        if self.charmodel_forward is not None:
-            char_reps_forward = self.build_char_reps(batch_forward_chars, batch_forward_offsets, device, forward=True)
+        if self.forward_charlm is not None:
+            char_reps_forward = self.build_char_reps(inputs, max_phrase_len, self.forward_charlm, self.charmodel_forward_projection, begin_paddings, device)
             all_inputs.append(char_reps_forward)
 
-        if self.charmodel_backward is not None:
-            char_reps_backward = self.build_char_reps(batch_backward_chars, batch_backward_offsets, device, forward=False)
+        if self.backward_charlm is not None:
+            char_reps_backward = self.build_char_reps(inputs, max_phrase_len, self.backward_charlm, self.charmodel_backward_projection, begin_paddings, device)
             all_inputs.append(char_reps_backward)
 
-        if len(all_inputs) > 1:
-            input_vectors = torch.cat(all_inputs, dim=2)
-        else:
-            input_vectors = all_inputs[0]
+        if self.config.use_elmo:
+            # this will be N arrays of 3xMx1024 where M is the number of words
+            # and N is the number of sentences (and 1024 is actually the number of weights)
+            elmo_arrays = self.elmo_model.sents2elmo(elmo_batch_words, output_layer=-2)
+            elmo_tensors = [torch.tensor(x).to(device=device) for x in elmo_arrays]
+            # elmo_tensor will now be Nx3xMx1024
+            elmo_tensor = torch.stack(elmo_tensors)
+            # Nx1024xMx3
+            elmo_tensor = torch.transpose(elmo_tensor, 1, 3)
+            # NxMx1024x3
+            elmo_tensor = torch.transpose(elmo_tensor, 1, 2)
+            # NxMx1024x1
+            elmo_tensor = self.elmo_combine_layers(elmo_tensor)
+            # NxMx1024
+            elmo_tensor = elmo_tensor.squeeze(3)
+            if self.config.elmo_projection:
+                elmo_tensor = self.elmo_projection(elmo_tensor)
+            all_inputs.append(elmo_tensor)
+
+        if self.bert_model is not None:
+            bert_embeddings = self.extract_bert_embeddings(inputs, max_phrase_len, begin_paddings, device)
+            all_inputs.append(bert_embeddings)
+
+        # still works even if there's just one item
+        input_vectors = torch.cat(all_inputs, dim=2)
+
+        if self.config.bilstm:
+            input_vectors, _ = self.bilstm(self.dropout(input_vectors))
 
         # reshape to fit the input tensors
         x = input_vectors.unsqueeze(1)
 
-        conv_outs = [self.dropout(F.relu(conv(x).squeeze(3)))
-                     for conv in self.conv_layers]
-        pool_outs = [F.max_pool1d(out, out.shape[2]).squeeze(2) for out in conv_outs]
+        conv_outs = []
+        for conv, filter_size in zip(self.conv_layers, self.config.filter_sizes):
+            if isinstance(filter_size, int):
+                conv_out = self.dropout(F.relu(conv(x).squeeze(3)))
+                conv_outs.append(conv_out)
+            else:
+                conv_out = conv(x).transpose(2, 3).flatten(1, 2)
+                conv_out = self.dropout(F.relu(conv_out))
+                conv_outs.append(conv_out)
+        pool_outs = [F.max_pool2d(out, (self.config.maxpool_width, out.shape[2])).squeeze(2) for out in conv_outs]
         pooled = torch.cat(pool_outs, dim=1)
 
         previous_layer = pooled
@@ -374,75 +449,11 @@ class CNNClassifier(nn.Module):
         return out
 
 
-# TODO: make some of the following methods part of the class
-
-# TODO: all this code is basically the same as for POS and NER.  Should refactor
-def save(filename, model, skip_modules=True):
-    model_state = model.state_dict()
-    # skip saving modules like pretrained embeddings, because they are large and will be saved in a separate file
-    if skip_modules:
-        skipped = [k for k in model_state.keys() if k.split('.')[0] in model.unsaved_modules]
-        for k in skipped:
-            del model_state[k]
-    params = {
-        'model': model_state,
-        'config': model.config,
-        'labels': model.labels,
-        'extra_vocab': model.extra_vocab,
-    }
-    try:
-        torch.save(params, filename, _use_new_zipfile_serialization=False)
-        logger.info("Model saved to {}".format(filename))
-    except (KeyboardInterrupt, SystemExit):
-        raise
-    except BaseException as e:
-        logger.warning("Saving failed to {}... continuing anyway.  Error: {}".format(filename, e))
-
-def load(filename, pretrain, charmodel_forward, charmodel_backward):
-    try:
-        checkpoint = torch.load(filename, lambda storage, loc: storage)
-    except BaseException:
-        logger.exception("Cannot load model from {}".format(filename))
-        raise
-    logger.debug("Loaded model {}".format(filename))
-
-    # TODO: should not be needed when all models have this value set
-    setattr(checkpoint['config'], 'char_lowercase', getattr(checkpoint['config'], 'char_lowercase', False))
-    setattr(checkpoint['config'], 'charlm_projection', getattr(checkpoint['config'], 'charlm_projection', None))
-
-    # TODO: the getattr is not needed when all models have this baked into the config
-    model_type = getattr(checkpoint['config'], 'model_type', 'CNNClassifier')
-    if model_type == 'CNNClassifier':
-        extra_vocab = checkpoint.get('extra_vocab', None)
-        model = CNNClassifier(pretrain=pretrain,
-                              extra_vocab=extra_vocab,
-                              labels=checkpoint['labels'],
-                              charmodel_forward=charmodel_forward,
-                              charmodel_backward=charmodel_backward,
-                              args=checkpoint['config'])
-    else:
-        raise ValueError("Unknown model type {}".format(model_type))
-    model.load_state_dict(checkpoint['model'], strict=False)
-
-    logger.debug("-- MODEL CONFIG --")
-    for k in model.config.__dict__:
-        logger.debug("  --{}: {}".format(k, model.config.__dict__[k]))
-
-    logger.debug("-- MODEL LABELS --")
-    logger.debug("  {}".format(" ".join(model.labels)))
-
-    return model
-
-
-def label_text(model, text, batch_size=None, reverse_label_map=None, device=None):
+def label_text(model, text, batch_size=None):
     """
     Given a list of sentences, return the model's results on that text.
     """
     model.eval()
-    if reverse_label_map is None:
-        reverse_label_map = {x: y for (x, y) in enumerate(model.labels)}
-    if device is None:
-        device = next(model.parameters()).device
 
     text = [data.update_text(s, model.config.wordvec_type) for s in text]
 
@@ -457,7 +468,7 @@ def label_text(model, text, batch_size=None, reverse_label_map=None, device=None
         if interval[1] - interval[0] == 0:
             # this can happen for empty text
             continue
-        output = model(text[interval[0]:interval[1]], device)
+        output = model(text[interval[0]:interval[1]])
         predicted = torch.argmax(output, dim=1)
         labels.extend(predicted.tolist())
 
