@@ -1,13 +1,18 @@
 """
 Utility functions.
 """
-import logging
-import os
+
 from collections import Counter
-import random
+from contextlib import contextmanager
+import gzip
 import json
+import logging
+import lzma
+import os
+import random
 import sys
 import unicodedata
+import zipfile
 
 import torch
 import numpy as np
@@ -46,6 +51,63 @@ def get_wordvec_file(wordvec_dir, shorthand, wordvec_type=None):
         filename = filename + ".txt"
     return filename
 
+@contextmanager
+def open_read_text(filename, encoding="utf-8"):
+    """
+    Opens a file as an .xz file or .gz if it ends with .xz or .gz, or regular text otherwise.
+
+    Use as a context
+
+    eg:
+    with open_read_text(filename) as fin:
+        do stuff
+
+    File will be closed once the context exits
+    """
+    if filename.endswith(".xz"):
+        with lzma.open(filename, mode='rt', encoding=encoding) as fin:
+            yield fin
+    elif filename.endswith(".gz"):
+        with gzip.open(filename, mode='rt', encoding=encoding) as fin:
+            yield fin
+    else:
+        with open(filename, encoding=encoding) as fin:
+            yield fin
+
+@contextmanager
+def open_read_binary(filename):
+    """
+    Opens a file as an .xz file or .gz if it ends with .xz or .gz, or regular binary file otherwise.
+
+    If a .zip file is given, it can be read if there is a single file in there
+
+    Use as a context
+
+    eg:
+    with open_read_binary(filename) as fin:
+        do stuff
+
+    File will be closed once the context exits
+    """
+    if filename.endswith(".xz"):
+        with lzma.open(filename, mode='rb') as fin:
+            yield fin
+    elif filename.endswith(".gz"):
+        with gzip.open(filename, mode='rb') as fin:
+            yield fin
+    elif filename.endswith(".zip"):
+        with zipfile.ZipFile(filename) as zin:
+            input_names = zin.namelist()
+            if len(input_names) == 0:
+                raise ValueError("Empty zip archive")
+            if len(input_names) > 1:
+                raise ValueError("zip file %s has more than one file in it")
+            with zin.open(input_names[0]) as fin:
+                yield fin
+    else:
+        with open(filename, mode='rb') as fin:
+            yield fin
+
 # training schedule
 def get_adaptive_eval_interval(cur_dev_size, thres_dev_size, base_interval):
     """ Adjust the evaluation interval adaptively.
@@ -118,7 +180,8 @@ def ensure_dir(d, verbose=True):
     if not os.path.exists(d):
         if verbose:
             logger.info("Directory {} does not exist; creating...".format(d))
-        os.makedirs(d)
+        # exist_ok: guard against race conditions
+        os.makedirs(d, exist_ok=True)
 
 def save_config(config, path, verbose=True):
     with open(path, 'w') as outfile:
@@ -282,6 +345,18 @@ def set_random_seed(seed, cuda):
         torch.cuda.manual_seed(seed)
     return seed
 
+def get_known_tags(known_tags):
+    """
+    Turns either a list or a list of lists into a single sorted list
+
+    Actually this is not at all necessarily about tags
+    """
+    if isinstance(known_tags, list) and isinstance(known_tags[0], list):
+        known_tags = sorted(set(x for y in known_tags for x in y))
+    else:
+        known_tags = sorted(known_tags)
+    return known_tags
+
 def find_missing_tags(known_tags, test_tags):
     if isinstance(known_tags, list) and isinstance(known_tags[0], list):
         known_tags = set(x for y in known_tags for x in y)
@@ -314,6 +389,9 @@ def get_tqdm():
       tqdm = utils.get_tqdm()
     then do this when you want a scroll bar or regular iterator depending on context:
       tqdm(list)
+
+    If there is no tty, the returned tqdm will always be disabled
+    unless disable=False is specifically set.
     """
     try:
         ipy_str = str(type(get_ipython()))
@@ -327,6 +405,37 @@ def get_tqdm():
         if sys.stderr.isatty():
             from tqdm import tqdm
             return tqdm
-    def tqdm(iterable, **kwargs):
-        return iterable
-    return tqdm
+
+    from tqdm import tqdm
+    def hidden_tqdm(*args, **kwargs):
+        if "disable" in kwargs:
+            return tqdm(*args, **kwargs)
+        kwargs["disable"] = True
+        return tqdm(*args, **kwargs)
+
+    return hidden_tqdm
+
+def checkpoint_name(save_dir, save_name, checkpoint_name):
+    """
+    Will return a recommended checkpoint name for the given dir, save_name, optional checkpoint_name
+
+    For example, can pass in args['save_dir'], args['save_name'], args['checkpoint_save_name']
+    """
+    if checkpoint_name:
+        return os.path.join(save_dir, checkpoint_name)
+
+    save_name = os.path.join(save_dir, save_name)
+    if save_name.endswith(".pt"):
+        return save_name[:-3] + "_checkpoint.pt"
+
+    return save_name + "_checkpoint"
+
+def load_elmo(elmo_model):
+    # This import is here so that Elmo integration can be treated
+    # as an optional feature
+    import elmoformanylangs
+
+    logger.info("Loading elmo: %s" % elmo_model)
+    elmo_model = elmoformanylangs.Embedder(elmo_model)
+    return elmo_model
+
