@@ -27,7 +27,7 @@ from stanza.utils.conll import CoNLL
 from stanza.models.common.doc import *
 from stanza.models import _training_logging
 
-from stanza.utils.confusion import format_confusion
+from stanza.utils.confusion import confusion_to_weighted_f1, format_confusion
 
 logger = logging.getLogger('stanza')
 
@@ -39,6 +39,7 @@ def parse_args(args=None):
     parser.add_argument('--wordvec_pretrain_file', type=str, default=None, help='Exact name of the pretrain file to read')
     parser.add_argument('--train_file', type=str, default=None, help='Input file for data loader.')
     parser.add_argument('--eval_file', type=str, default=None, help='Input file for data loader.')
+    parser.add_argument('--eval_output_file', type=str, default=None, help='Where to write results: text, gold, pred.  If None, no results file printed')
 
     parser.add_argument('--mode', default='train', choices=['train', 'predict'])
     parser.add_argument('--finetune', action='store_true', help='Load existing model during `train` mode from `save_dir` path')
@@ -55,7 +56,7 @@ def parse_args(args=None):
     parser.add_argument('--num_layers', type=int, default=1)
     parser.add_argument('--char_num_layers', type=int, default=1)
     parser.add_argument('--pretrain_max_vocab', type=int, default=100000)
-    parser.add_argument('--word_dropout', type=float, default=0)
+    parser.add_argument('--word_dropout', type=float, default=0.01, help="How often to remove a word at training time.  Set to a small value to train unk when finetuning word embeddings")
     parser.add_argument('--locked_dropout', type=float, default=0.0)
     parser.add_argument('--dropout', type=float, default=0.5)
     parser.add_argument('--rec_dropout', type=float, default=0, help="Word recurrent dropout")
@@ -70,9 +71,14 @@ def parse_args(args=None):
     parser.add_argument('--char_lowercase', dest='char_lowercase', action='store_true', help="Use lowercased characters in character model.")
     parser.add_argument('--no_lowercase', dest='lowercase', action='store_false', help="Use cased word vectors.")
     parser.add_argument('--no_emb_finetune', dest='emb_finetune', action='store_false', help="Turn off finetuning of the embedding matrix.")
+    parser.add_argument('--emb_finetune_known_only', dest='emb_finetune_known_only', action='store_true', help="Finetune the embedding matrix only for words in the embedding.  (Default: finetune words not in the embedding as well)  This may be useful for very large datasets where obscure words are only trained once in a while, such as French-WikiNER")
     parser.add_argument('--no_input_transform', dest='input_transform', action='store_false', help="Do not use input transformation layer before tagger lstm.")
     parser.add_argument('--scheme', type=str, default='bioes', help="The tagging scheme to use: bio or bioes.")
 
+
+    parser.add_argument('--bert_model', type=str, default=None, help="Use an external bert model (requires the transformers package)")
+    parser.add_argument('--no_bert_model', dest='bert_model', action="store_const", const=None, help="Don't use bert")
+
     parser.add_argument('--sample_train', type=float, default=1.0, help='Subsample training data.')
     parser.add_argument('--optim', type=str, default='sgd', help='sgd, adagrad, adam or adamax.')
     parser.add_argument('--lr', type=float, default=0.1, help='Learning rate.')
@@ -93,17 +99,24 @@ def parse_args(args=None):
     parser.add_argument('--cuda', type=bool, default=torch.cuda.is_available())
     parser.add_argument('--cpu', action='store_true', help='Ignore CUDA.')
 
+    parser.add_argument('--wandb', action='store_true', help='Start a wandb session and write the results of training.  Only applies to training.  Use --wandb_name instead to specify a name')
+    parser.add_argument('--wandb_name', default=None, help='Name of a wandb session to start when training.  Will default to the dataset short name')
+
     args = parser.parse_args(args=args)
+
+    if args.wandb_name:
+        args.wandb = True
+    if args.cpu:
+        args.cuda = False
+
+    args = vars(args)
     return args
 
 def main(args=None):
     args = parse_args(args=args)
 
-    if args.cpu:
-        args.cuda = False
-    utils.set_random_seed(args.seed, args.cuda)
+    utils.set_random_seed(args['seed'], args['cuda'])
 
-    args = vars(args)
     logger.info("Running NER tagger in {} mode".format(args['mode']))
 
     if args['mode'] == 'train':
@@ -111,9 +124,23 @@ def main(args=None):
     else:
         evaluate(args)
 
+def load_pretrain(args):
+    # load pretrained vectors
+    if args['wordvec_pretrain_file']:
+        pretrain_file = args['wordvec_pretrain_file']
+        pretrain = Pretrain(pretrain_file, None, args['pretrain_max_vocab'], save_to_file=False)
+    else:
+        if len(args['wordvec_file']) == 0:
+            vec_file = utils.get_wordvec_file(args['wordvec_dir'], args['shorthand'])
+        else:
+            vec_file = args['wordvec_file']
+        # do not save pretrained embeddings individually
+        pretrain = Pretrain(None, vec_file, args['pretrain_max_vocab'], save_to_file=False)
+    return pretrain
+
 def train(args):
     utils.ensure_dir(args['save_dir'])
-    model_file = os.path.join(args['save_dir'], args['save_name']) if args['save_name'] is not None \
+    model_file = os.path.join(args['save_dir'], args['save_name']) if args['save_name'] \
         else '{}/{}_nertagger.pt'.format(args['save_dir'], args['shorthand'])
 
     pretrain = None
@@ -130,17 +157,7 @@ def train(args):
         if args['finetune']:
             raise FileNotFoundError('Finetune is set to true but model file is not found: {}'.format(model_file))
 
-        # load pretrained vectors
-        if args['wordvec_pretrain_file']:
-            pretrain_file = args['wordvec_pretrain_file']
-            pretrain = Pretrain(pretrain_file, None, args['pretrain_max_vocab'], save_to_file=False)
-        else:
-            if len(args['wordvec_file']) == 0:
-                vec_file = utils.get_wordvec_file(args['wordvec_dir'], args['shorthand'])
-            else:
-                vec_file = args['wordvec_file']
-            # do not save pretrained embeddings individually
-            pretrain = Pretrain(None, vec_file, args['pretrain_max_vocab'], save_to_file=False)
+        pretrain = load_pretrain(args)
 
         if pretrain is not None:
             word_emb_dim = pretrain.emb.shape[1]
@@ -160,12 +177,23 @@ def train(args):
     # load data
     logger.info("Loading data with batch size {}...".format(args['batch_size']))
     train_doc = Document(json.load(open(args['train_file'])))
+    logger.info("Loaded %d sentences of training data", len(train_doc.sentences))
+    if len(train_doc.sentences) == 0:
+        raise ValueError("File %s exists but has no usable training data" % args['train_file'])
     train_batch = DataLoader(train_doc, args['batch_size'], args, pretrain, vocab=vocab, evaluation=False)
     vocab = train_batch.vocab
     dev_doc = Document(json.load(open(args['eval_file'])))
+    logger.info("Loaded %d sentences of dev data", len(dev_doc.sentences))
+    if len(dev_doc.sentences) == 0:
+        raise ValueError("File %s exists but has no usable dev data" % args['train_file'])
     dev_batch = DataLoader(dev_doc, args['batch_size'], args, pretrain, vocab=vocab, evaluation=True)
     dev_gold_tags = dev_batch.tags
 
+    train_tags = utils.get_known_tags(train_batch.tags)
+    logger.info("Tags present in training set:\n  Tags without BIES markers: %s\n  Tags with B-, I-, E-, or S-: %s",
+                " ".join(sorted(set(i for i in train_tags if i[:2] not in ('B-', 'I-', 'E-', 'S-')))),
+                " ".join(sorted(set(i[2:] for i in train_tags if i[:2] in ('B-', 'I-', 'E-', 'S-')))))
+
     if args['finetune']:
         utils.warn_missing_tags([i for i in trainer.vocab['tag']], train_batch.tags, "training set")
     utils.warn_missing_tags(train_batch.tags, dev_batch.tags, "dev set")
@@ -173,7 +201,7 @@ def train(args):
     # skip training if the language does not have training or dev data
     if len(train_batch) == 0 or len(dev_batch) == 0:
         logger.info("Skip training because no data available...")
-        sys.exit(0)
+        return
 
     logger.info("Training tagger...")
     if trainer is None: # init if model was not loaded previously from file
@@ -196,6 +224,13 @@ def train(args):
     else:
         scheduler = None
 
+    if args['wandb']:
+        import wandb
+        wandb_name = args['wandb_name'] if args['wandb_name'] else "%s_ner" % args['shorthand']
+        wandb.init(name=wandb_name, config=args)
+        wandb.run.define_metric('train_loss', summary='min')
+        wandb.run.define_metric('dev_score', summary='max')
+
     # start training
     train_loss = 0
     while True:
@@ -221,6 +256,8 @@ def train(args):
 
                 train_loss = train_loss / args['eval_interval'] # avg loss per batch
                 logger.info("step {}: train_loss = {:.6f}, dev_score = {:.4f}".format(global_step, train_loss, dev_score))
+                if args['wandb']:
+                    wandb.log({'train_loss': train_loss, 'dev_score': dev_score})
                 train_loss = 0
 
                 # save best model
@@ -249,21 +286,46 @@ def train(args):
 
     logger.info("Training ended with {} steps.".format(global_step))
 
+    if args['wandb']:
+        wandb.finish()
+
     if len(dev_score_history) > 0:
         best_f, best_eval = max(dev_score_history)*100, np.argmax(dev_score_history)+1
         logger.info("Best dev F1 = {:.2f}, at iteration = {}".format(best_f, best_eval * args['eval_interval']))
+    else:
+        logger.info("Dev set never evaluated.  Saving final model.")
+        trainer.save(model_file)
+
+def write_ner_results(filename, batch, preds):
+    if len(batch.tags) != len(preds):
+        raise ValueError("Unexpected batch vs pred lengths: %d vs %d" % (len(batch.tags), len(preds)))
+
+    with open(filename, "w", encoding="utf-8") as fout:
+        tag_idx = 0
+        for b in batch:
+            # b[0] is words, b[5] is orig_idx
+            # a namedtuple would make this cleaner without being much slower
+            text = utils.unsort(b[0], b[5])
+            for sentence in text:
+                sentence_gold = batch.tags[tag_idx]
+                sentence_pred = preds[tag_idx]
+                tag_idx += 1
+                for word, gold, pred in zip(sentence, sentence_gold, sentence_pred):
+                    fout.write("%s\t%s\t%s\n" % (word, gold, pred))
+                fout.write("\n")
 
 def evaluate(args):
     # file paths
-    model_file = os.path.join(args['save_dir'], args['save_name']) if args['save_name'] is not None \
+    model_file = os.path.join(args['save_dir'], args['save_name']) if args['save_name'] \
         else '{}/{}_nertagger.pt'.format(args['save_dir'], args['shorthand'])
 
     loaded_args, trainer, vocab = load_model(args, model_file)
+    logger.debug("Loaded model for eval from %s", model_file)
 
     # load data
     logger.info("Loading data with batch size {}...".format(args['batch_size']))
     doc = Document(json.load(open(args['eval_file'])))
-    batch = DataLoader(doc, args['batch_size'], loaded_args, vocab=vocab, evaluation=True)
+    batch = DataLoader(doc, args['batch_size'], loaded_args, vocab=vocab, evaluation=True, bert_tokenizer=trainer.bert_tokenizer)
     utils.warn_missing_tags([i for i in trainer.vocab['tag']], batch.tags, "eval_file")
 
     logger.info("Start evaluation...")
@@ -274,11 +336,16 @@ def evaluate(args):
     gold_tags = batch.tags
     _, _, score = scorer.score_by_entity(preds, gold_tags)
     _, _, _, confusion = scorer.score_by_token(preds, gold_tags)
+    logger.info("Weighted f1 for non-O tokens: %5f", confusion_to_weighted_f1(confusion, exclude=["O"]))
 
     logger.info("NER tagger score:")
     logger.info("{} {:.2f}".format(args['shorthand'], score*100))
     logger.info("NER token confusion matrix:\n{}".format(format_confusion(confusion)))
 
+    if args['eval_output_file']:
+        write_ner_results(args['eval_output_file'], batch, preds)
+
+    return confusion
 
 def load_model(args, model_file):
     # load model
@@ -288,7 +355,8 @@ def load_model(args, model_file):
         charlm_args['charlm_forward_file'] = args['charlm_forward_file']
     if 'charlm_backward_file' in args:
         charlm_args['charlm_backward_file'] = args['charlm_backward_file']
-    trainer = Trainer(args=charlm_args, model_file=model_file, use_cuda=use_cuda, train_classifier_only=args['train_classifier_only'])
+    pretrain = load_pretrain(args)
+    trainer = Trainer(args=charlm_args, model_file=model_file, pretrain=pretrain, use_cuda=use_cuda, train_classifier_only=args['train_classifier_only'])
     loaded_args, vocab = trainer.args, trainer.vocab
 
     # load config
