@@ -4,8 +4,8 @@ from stanza.models.common.vocab import BaseVocab, BaseMultiVocab, CharVocab
 from stanza.models.common.vocab import CompositeVocab, VOCAB_PREFIX, EMPTY, EMPTY_ID
 
 class WordVocab(BaseVocab):
-    def __init__(self, data=None, lang="", idx=0, cutoff=0, lower=False, ignore=[]):
-        self.ignore = ignore
+    def __init__(self, data=None, lang="", idx=0, cutoff=0, lower=False, ignore=None):
+        self.ignore = ignore if ignore is not None else []
         super().__init__(data, lang=lang, idx=idx, cutoff=cutoff, lower=lower)
         self.state_attrs += ['ignore']
 
@@ -55,9 +55,9 @@ class MultiVocab(BaseMultiVocab):
     @classmethod
     def load_state_dict(cls, state_dict):
         class_dict = {'CharVocab': CharVocab,
-                'WordVocab': WordVocab,
-                'XPOSVocab': XPOSVocab,
-                'FeatureVocab': FeatureVocab}
+                      'WordVocab': WordVocab,
+                      'XPOSVocab': XPOSVocab,
+                      'FeatureVocab': FeatureVocab}
         new = cls()
         assert '_key2class' in state_dict, "Cannot find class name mapping in state dict!"
         key2class = state_dict.pop('_key2class')
